#include "IR.hpp"
#include "AST.hpp"
#include "GRCPrinter.hpp"
#include "CPrinter.hpp"
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>

using namespace AST;
using namespace std;
//using namespace GRCDot;

class ModifyGraph
{
  set<GRCNode *> visited;
  GRCNode *top; //top of the flow graph

public:
  ModifyGraph(GRCNode *top):top(top)
  {
    EnterGRC *et = dynamic_cast<EnterGRC *>(top);
    assert(et);

    ReplaceNull2Nop(top);

    visited.clear();
    DFS(top);
  }

  GRCNode *addExit()
  {
    set<GRCNode *>::iterator i;
    set<GRCNode *> end_nodes;
    GRCNode *exitgrc = new ExitGRC();
    
    for (i = visited.begin(); i != visited.end(); i++){
      
      if ((*i)->successors.size() == 0){
	if (((*i)->predecessors.size() > 0) || (dynamic_cast<EnterGRC *>(*i)))
	  end_nodes.insert(*i);
      }
    }
    
    for (i = end_nodes.begin(); i != end_nodes.end(); i++){
      if (dynamic_cast<Nop *>(*i)){
	assert((*i)->predecessors.size() == 1);
	exitgrc->predecessors.push_back((*i)->predecessors[0]);
	vector<GRCNode *>::iterator j;
	GRCNode *par = (*i)->predecessors[0];
	for (j = par->successors.begin(); j != par->successors.end(); j++){
	  if (*j == *i){
	    *j = exitgrc;
	    break;
	  }
	}
      }
      else{
	(*i)->successors.push_back(exitgrc);
	exitgrc->predecessors.push_back(*i);
      }
    }
    
    return exitgrc;
  }
  
  GRCNode *removeEnter()
  {   
    GRCNode *org_top = top;
    assert(top->successors.size()>0);
    top = top->successors[0];
    assert(top != org_top);
    org_top->successors.clear();    
    top->predecessors.clear();
    assert(top->successors.size()>0);
    return top;
  }
  
private:
  void ReplaceNull2Nop(GRCNode *n)
  {
    if (visited.find(n) != visited.end())
      return;
    visited.insert(n);
    
    vector<GRCNode *>::iterator i;
    for (i = n->successors.begin(); i != n->successors.end(); i++)
      if (!*i){
	*i = new Nop();
	(*i)->predecessors.push_back(n);
	visited.insert(*i);
      }
      else
	ReplaceNull2Nop(*i);
  }

  void DFS(GRCNode *n)
  {
    assert(n);
    if (visited.find(n) != visited.end())
      return;
    visited.insert(n);
    vector<GRCNode *>::iterator i;
    for (i = n->successors.begin(); i != n->successors.end(); i++)
      DFS(*i);
  }

};

class CCodeGenerator
{
public:
  CCodeGenerator()
  {

  }

  void excute(ofstream &o, Module &m, GRCNode *exitgrc)
  {
    CPrinter::Printer p(o, m);

    GRCgraph *g = dynamic_cast<AST::GRCgraph*>(m.body);
    
    if (!g) throw IR::Error("Module is not in GRC format");

    // FIXME: Should add a command-line argument to allow the user to
    // override this
    string basename = m.symbol->name;
    
    p.printDeclarations(basename);
    p.ioDefinitions();

    o <<
      "\n"
      "int " << m.symbol->name << "(void)\n"
      "{\n"
      ;


    for (map<Sync *, string>::iterator i = p.terminationVar.begin();
	 i != p.terminationVar.end(); i++)
      o << "\t"<<(*i).second<<"=-1;\n";

    // FIXME: insert the code that generates the body of the function here
    p.printStructuredCode(exitgrc);

    // Call an emit function for each emitted signal and reset each input

    p.outputFunctions();
    p.resetInputs();

    STexcl *excl = dynamic_cast<STexcl*>(g->selection_tree);
    assert(excl);
    assert(!excl->children.empty());
    
    STleaf *terminal_leaf = dynamic_cast<STleaf*>(excl->children.front());
    
    if (terminal_leaf && terminal_leaf->isfinal()) {
      o << "  return (" << p.stateVar[excl] << " != 0);\n";
    } else {
      o << "  return 1;\n";
    }
    o <<
      "}\n";
    
    // Reset function
    
    o <<
      "\n"
      "int " << m.symbol->name << "_reset(void)\n"
      "{\n";

    p.resetInputs();
    
    o <<
    "  return 0;\n"
      "}\n";
  }

};

int main(int argc, char *argv[])
{

  if (argc < 3) {
    cerr<<"Usage: ... scfg_file c_code_file\n"; return -1;}

  ifstream scfgf(argv[1]); if(!scfgf.good()) return -1;
  ofstream ccodef(argv[2]);if(!ccodef.good()) return -1;

  try {
    IR::XMListream r(scfgf);
    IR::Node *n;
    r >> n;
    scfgf.close();
    AST::Modules *mods = dynamic_cast<AST::Modules*>(n);
    if (!mods) throw IR::Error("Root node is not a Modules object");

    assert(mods->modules.size()>0);
    GRCgraph *gf = dynamic_cast<GRCgraph*>((*(mods->modules.begin()))->body);
    assert(gf);  
    GRCNode *top = gf->control_flow_graph;
    
    ModifyGraph *mg = new ModifyGraph(top);
    
    //add ExitGRC
    GRCNode *exitgrc = mg->addExit();
    assert(exitgrc);

    //remove EnterGRC
    gf->control_flow_graph = mg->removeEnter();
    assert(gf->control_flow_graph->successors.size()>0);

    //c code generating
    AST::Module *m = mods->modules.front();
    assert(m);
    CCodeGenerator *cg = new CCodeGenerator();
    cg->excute(ccodef, *m, exitgrc);

    ccodef.close();
  } catch (IR::Error &e) {
    cerr << e.s << endl;
    exit(-1);
  }

  return 0;

}
